//	GeometryGamesExportView.swift
//
//	© 2022 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


struct GeometryGamesExportView<ModelDataType>: View {

	enum ExportedImageSize: Int, CaseIterable, Identifiable {

		case size256  =  256
		case size512  =  512
		case size1024 = 1024
		case size2048 = 2048
		case size4096 = 4096
		
		var id: Int { self.rawValue }
		
		var asString: String {
			switch self {
			case .size256:
				return "256 × 256"
			case .size512:
				return "512 × 512"
			case .size1024:
				return "1024 × 1024"
			case .size2048:
				return "2048 × 2048"
			case .size4096:
				return "4096 × 4096"
			}
		}
	}

	@State var imageSizeInPixels = ExportedImageSize.size256	//	to be overridden in onAppear below
	@StateObject var itsExporter = GeometryGamesDocumentExporter<ModelDataType>()
	let theDefaultExportedImageSize = ExportedImageSize.size1024
	let itsModelData: ModelDataType
	let itsExportRenderer: GeometryGamesRenderer<ModelDataType>
	let itsSourceDocumentURL: URL?	//	KaleidoPaint and 4D Draw pass the current drawing's file URL
									//		so that we may save exported images
									//		into a directory whose name matches the drawing's filename
									//		(but of course without the filename extension).

	init(
		modelData: ModelDataType,
		exportRenderer: GeometryGamesRenderer<ModelDataType>,
		sourceDocumentURL: URL? = nil	//	See comment accompanying itsSourceDocumentURL above.
	) {
		itsModelData = modelData
		itsExportRenderer = exportRenderer
		itsSourceDocumentURL = sourceDocumentURL
	}
	
	var body: some View {
	
		let theButtonPadding: CGFloat = 4
		
		let theImageSizeString = Bundle.main.localizedString(
			forKey: "Image Size",
			value: nil,
			table: "GeometryGamesLocalizable")

		let theSaveImageString = Bundle.main.localizedString(
			forKey: "Save Image",
			value: nil,
			table: "GeometryGamesLocalizable")

		let theCopyImageString = Bundle.main.localizedString(
			forKey: "Copy Image",
			value: nil,
			table: "GeometryGamesLocalizable")
		
		return VStack(alignment: .center) {
			
			HStack() {

//	The Picker worked fine when running on an iOS device,
//	but didn't work at all when running on a Mac.
//
//				Picker(
//					theImageSizeString,
//					selection: $imageSizeInPixels
//				) {
//					ForEach(ExportedImageSize.allCases, id: \.id) {theImageSize in
//						Text(verbatim: theImageSize.asString).tag(theImageSize)
//					}
//				}
//				.pickerStyle(MenuPickerStyle())
//
//	So instead let's let a Menu contain the Picker, which should be just as good.
//	Well, on macOS it inserts an extra level in the menu heirarchy, that is,
//
//		Image Size > Image Size > {options}
//
//	instead of just
//
//		Image Size > {options}
//
//	as on iOS, but I guess we can live with that.
//
				Menu() {
					Picker(selection: $imageSizeInPixels, label: Text(theImageSizeString)) {
						ForEach(ExportedImageSize.allCases, id: \.id) {theImageSize in
							Text(theImageSize.asString)
								.tag(theImageSize)
						}
					}
				} label: {
					Text(theImageSizeString)
				}

				Spacer()

				Text(formattedImageSize(imageSizeInPixels.rawValue))
					.foregroundColor(.gray)
			}
			
			//	Caution:  Divider() will make the HStack fill all available width
			//	unless we apply .fixedSize() or .fixedSize(horizontal: vertical: )
			//	to the containing VStack().
			Divider()
		
			Button() {
				let (_, _) = itsExporter.saveImage(
					modelData: itsModelData,
					renderer: itsExportRenderer,
					requestedWidthPx:  imageSizeInPixels.rawValue,
					requestedHeightPx: imageSizeInPixels.rawValue,
					sourceDocumentURL: itsSourceDocumentURL)
			} label: {
				HStack() {
					Text(theSaveImageString)
					Spacer()
					Image(systemName: "folder")
				}
				.padding(EdgeInsets(top: theButtonPadding, leading: 0, bottom: theButtonPadding, trailing: 0))
			}

			Button {
				let (_, _) = itsExporter.copyImage(
					modelData: itsModelData,
					renderer: itsExportRenderer,
					requestedWidthPx:  imageSizeInPixels.rawValue,
					requestedHeightPx: imageSizeInPixels.rawValue)
			} label: {
			
				HStack() {
					Text(theCopyImageString)
					Spacer()
					Image(systemName: "doc.on.doc")
				}
				.padding(EdgeInsets(top: theButtonPadding, leading: 0, bottom: theButtonPadding, trailing: 0))
			}
		}
		.fixedSize()
		.onAppear() {
			imageSizeInPixels = ExportedImageSize(rawValue:
				UserDefaults.standard.object(forKey: "exported image size")
					as? Int ?? theDefaultExportedImageSize.rawValue
			) ?? theDefaultExportedImageSize
		}
		.onDisappear() {
			UserDefaults.standard.setValue(
				imageSizeInPixels.rawValue,
				forKey: "exported image size")
		}
		.padding(geometryGamesPanelPadding)
		.background(Color.white)
		.cornerRadius(geometryGamesCornerRadius)
	}
	
	func formattedImageSize(_ size: Int) -> String {
		
		let theNumberFormatter = NumberFormatter()
		theNumberFormatter.groupingSeparator = ""	//	suppress the comma in "1,024" etc.

		let theFormattedNumber = (theNumberFormatter.string(from: NSNumber(value: size))
									?? "error")
		
		let theDimensions = theFormattedNumber + " × " + theFormattedNumber
		
		return theDimensions
	}
}


class GeometryGamesDocumentExporter<ModelDataType>: NSObject, UIDocumentPickerDelegate, ObservableObject {
	//	To conform to UIDocumentPickerDelegate, we must also inherit from NSObject.
	
	func copyImage(
		modelData: ModelDataType,
		renderer: GeometryGamesRenderer<ModelDataType>,
		requestedWidthPx: Int,
		requestedHeightPx: Int
	) -> (Int, Int) {	//	returns rendered size

		let (theOptionalPngRepresentation, theRenderedWidthPx, theRenderedHeightPx)
			= renderer.offscreenImageAsPNG(
				modelData: modelData,
				requestedWidthPx: requestedWidthPx,
				requestedHeightPx: requestedHeightPx)

		guard let thePngRepresentation = theOptionalPngRepresentation else {
			return (0, 0)
		}

		UIPasteboard.general.setData(thePngRepresentation, forPasteboardType: "public.png")

		return (theRenderedWidthPx, theRenderedHeightPx)
	}

	func saveImage(
		modelData: ModelDataType,
		renderer: GeometryGamesRenderer<ModelDataType>,
		requestedWidthPx: Int,
		requestedHeightPx: Int,
		sourceDocumentURL: URL?	//	within the app's Documents folder
	) -> (Int, Int) {	//	returns rendered size

		let (theOptionalPngRepresentation, theRenderedWidthPx, theRenderedHeightPx)
			= renderer.offscreenImageAsPNG(
				modelData: modelData,
				requestedWidthPx: requestedWidthPx,
				requestedHeightPx: requestedHeightPx)

		guard let thePngRepresentation = theOptionalPngRepresentation else {
			return (0, 0)
		}

		let (_, theTemporaryImageURL) = temporaryFileURL()

		//	Note:  The following code uses UIKit.
		//	Maybe some future version of iOS will let us
		//	use SwiftUI code for this purpose instead.
		do {
			try thePngRepresentation.write(to: theTemporaryImageURL)

			//	In principle we'd be happy to let
			//	iOS move the temporary file,
			//	with no need to make a copy of it.
			//	The downside of that approach, though,
			//	is that the Document Picker would invite
			//	the user to "Move" the exported image.
			//	We'd much prefer that the Document Picker
			//	invite the user to "Save" the exported image,
			//	and for that to happen we need to ask it
			//	to save a copy of the temporary file.
			let theDocumentPickerViewController = UIDocumentPickerViewController(
													forExporting: [theTemporaryImageURL],
													asCopy: true)

			//	In principle we could ignore the temporary file,
			//	and iOS would eventually delete it (because it's
			//	in the temporaryDirectory).  But a cleaner solution
			//	is to make ourselves theDocumentPickerViewController's delegate
			//	and let the delegate callback methods delete
			//	the temporary file immediately.
			theDocumentPickerViewController.delegate = self

			//	If the caller asked that we start the user
			//	in a given default directory, create that directory (if needed)
			//	and pass its URL to theDocumentPickerViewController.
			//	The user is, of course, free to navigate
			//	to some other directory if s/he wishes.
			if let theSourceDocumentURL = sourceDocumentURL {

				let theFileManger = FileManager.default

				let theSubfolderURL = theSourceDocumentURL.deletingPathExtension()

				try theFileManger.createDirectory(	//	It's OK if the directory already exists.
					at: theSubfolderURL,
					withIntermediateDirectories: true,
					attributes: nil)

				theDocumentPickerViewController.directoryURL = theSubfolderURL
			}

			//	Locate top UIViewController on the stack,
			//	and ask it to present theDocumentPickerViewController.
			//
			//		Technical flaw
			//
			//			I would have much rather passed a reference
			//			to the GeometryGamesExportView into this
			//			GeometryGamesDocumentExporter, and used it
			//			to get a reference to its underlying UIViewController.
			//			But I can't figure out how to get from a SwiftUI View
			//			to a UIView or UIViewController.  So as a workaround,
			//			the code below fishes up an arbitrary window and uses it
			//			instead, which works in practice, even though it's
			//			a horrible "code smell".
			//
			//			Maybe in the future SwiftUI will drop its dependence
			//			on UIKit, and there'll be some cleaner way to present
			//			a purely SwiftUI DocumentPicker (keeping in mind that
			//			we're not creating a document-based app, we're just
			//			exporting an image from a non-document-based app).
			//
			if let theWindowScene = UIApplication.shared.connectedScenes.first as? UIWindowScene {
				if let theRootViewController = theWindowScene.windows.first?.rootViewController {

					var theTopViewController = theRootViewController
					while let thePresentedController = theTopViewController.presentedViewController {
						theTopViewController = thePresentedController
					}
		
					theTopViewController.present(theDocumentPickerViewController, animated: false)
				}
			}
			
		} catch {
			assertionFailure("Error saving image file")
		}

		return (theRenderedWidthPx, theRenderedHeightPx)
	}

	func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL]) {
		deleteTemporaryFile()
	}

	func documentPickerWasCancelled(_ controller: UIDocumentPickerViewController) {
		deleteTemporaryFile()
	}

	func deleteTemporaryFile() {

		let (theFileManager, theTemporaryImageURL) = temporaryFileURL()
		do {
			try theFileManager.removeItem(at: theTemporaryImageURL)
		} catch {
			assertionFailure("Error deleting temporary image file")
		}
	}

	func temporaryFileURL() -> (FileManager, URL) {

		let theFileManager = FileManager.default
		let theTemporaryFileName = Bundle.main.localizedString(
										forKey: "Untitled",
										value: nil,
										table: "GeometryGamesLocalizable"
									).appending(".png")
		let theTemporaryFileURL = theFileManager
									.temporaryDirectory
									.appendingPathComponent(theTemporaryFileName)

		return (theFileManager, theTemporaryFileURL)
	}
}
