//	GeometryGamesView.swift
//
//	© 2022 by Jeff Weeks
//	See TermsOfUse.txt

import UIKit
import Metal
import CoreMotion

protocol ModelIsAnimatable {
	func updateModel() -> UInt64	//	returns the model's display change count
}

class GeometryGamesView<ModelDataType: ModelIsAnimatable>: UIView {

	let itsModelData: ModelDataType
	let itsRenderer: GeometryGamesRenderer<ModelDataType>

	//	itsAnimationTimer keeps a strong reference to its target,
	//	namely this GeometryGamesView.  If this GeometryGamesView
	//	also kept a strong reference to itsAnimationTimer,
	//	we'd have a strong reference cycle and neither object
	//	would ever get deallocated.  So keep a weak reference instead.
	//	The run loop will keep a strong reference to itsAnimationTimer
	//	as long as its valid, so there's no danger
	//	of it getting deallocated too soon.
	//
	//		Note:  Because itsAnimationTimer retains this GeometryGamesView,
	//		the GeometryGamesView would never get deallocated
	//		if we merely paused itsAnimationTimer when not needed.
	//		Instead we must invalidate (and therefore deallocate)
	//		the timer when not needed, and recreate it when needed again.
	//
	weak var itsAnimationTimer: CADisplayLink? = nil
		
	//	When the view size changes, we manually adjust
	//	the drawable size accordingly, and then need
	//	to request a redraw.
	var itsDrawableSizeHasChanged: Bool = false
	
	//	
	//	What was the value of the ModelData's change count
	//	the last time we refreshed our view(s)?
	//
	//		Note:  itsModelData initializes its change count to 0,
	//		so we should initialize itsPreviousModelChangeCount
	//		to a different value to ensure an initial update.
	//
	var itsPreviousModelChangeCount: UInt64 = 0xFFFFFFFFFFFFFFFF
	
	//	A DisplayPoint is convenient for report touch locations
	//	in a simple format.
	struct DisplayPoint {
	
		//	Measurements may be in pixels or points;
		//	either is fine just so they're consistent.
	
		var x: Double	//	left-to-right, from 0 to viewWidth
		var y: Double	//	bottom-to-top, from 0 to viewHeight
		var viewWidth: Double
		var viewHeight: Double
	}

	//	A DisplayPointMotion is just like a DisplayPoint, except that
	//	it stores a relative motion instead of an absolute position.
	struct DisplayPointMotion {
	
		//	Measurements may be in pixels or points;
		//	either is fine just so they're consistent.
	
		var deltaX: Double	//	left-to-right motion is positive
		var deltaY: Double	//	bottom-to-top motion is positive
		var viewWidth: Double
		var viewHeight: Double
	}


	override class var layerClass: AnyClass {
		return CAMetalLayer.self}


	init(
		frame frameRect: CGRect,
		modelData: ModelDataType,
		renderer: GeometryGamesRenderer<ModelDataType>
	) {
	
		itsModelData = modelData
		itsRenderer = renderer
		
		//	Call super.init().
		//	Among other things, super.init() will create the CAMetalLayer.
		super.init(frame: frameRect)

		//	super.init() has created the CAMetalLayer,
		//	so we may now set the CAMetalLayer's pixel format.
		//
		//		Note:  The pixel format here must agree
		//		with the itsRenderer's pixel format.
		//
		if let theCAMetalLayer = self.layer as? CAMetalLayer {
			theCAMetalLayer.device = renderer.itsDevice
			theCAMetalLayer.pixelFormat = renderer.itsColorPixelFormat
		}

		//	We'll draw all Metal graphics at the display's native scale.
		//	Please see the comments in layoutSubviews() for further details.
		//	See also Apple's Technical Q&A QA1909
		//
		//			https://developer.apple.com/library/content/qa/qa1909/_index.html
		//
		self.contentScaleFactor = UIScreen.main.nativeScale

		//	Create itsAnimationTimer later, in didMoveToWindow().
		itsAnimationTimer = nil	//	redundant, but makes our intentions clear
	}

	required init(coder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}
	
	//	layoutSubviews() is really a "viewSizeDidChange()" message.
	override func layoutSubviews() {
	
		//	The view size has changed, so we must manually
		//	adjust the CAMetalLayer's drawableSize accordingly.

		//	The tricky part is that iOS might be drawing
		//	the user interface elements at some integer scale
		//	(either 2x or 3x) and then resampling the result
		//	at some non-integer scale to fit the device's display.
		//	For example, the iPhone 6+, 6s+, 7+ and 8+ all normally
		//	draw at 3x and then downsample to ~2.6x (1080/414).
		//	Some of the other iPhones, such as my iPhone SE 2020,
		//	normally render at 2x, but also offer a Display Zoom mode
		//	in which a 2x image gets upsampled to ~2.3x (750/320).
		//
		//	Happily, iOS let's us bypass that resampling,
		//	and instead render our Metal image at the device's
		//	native pixel resolution.  To do so, we need to set
		//	theDrawableSize to exactly the size of the area that's
		//	been assigned to our view (measured in native pixels).
		//
		//	Unhappily, iOS doesn't tell us what that required
		//	size (measured in native pixels) is.  Instead, we must
		//	figure it out for ourselves, by computing
		//
		//		theDrawableSize (in pixels)
		//		   = the view size (in points)
		//		   * the content scale factor (in pixels/point)
		//
		//	So far so good.
		//	The only catch is that the computed value of theDrawableSize
		//	typically comes out to a non-integer, and if we round it
		//	in the wrong direction to get an integer, our value
		//	for theDrawableSize might be 1 pixel wider or narrower
		//	(and taller or shorter) than the area iOS has assigned
		//	to our view, in which case iOS might end up re-sampling
		//	our image after all -- which is something we were hoping
		//	to avoid.
		//
		//	In practice, we can pass our computed non-integer value
		//	of theDrawableSize and let iOS do the rounding
		//	(which it does immediately, apparently in the "setter"
		//	for theCAMetalLayer.drawableSize).  Moreover, testing shows
		//	that iOS always rounds down, so for example if we request
		//	a width of 636.87, we'll get 636, not 637.
		//
		//	In most cases, passing the un-rounded value should
		//	let us match the size of the assigned area (in native pixels)
		//	exactly.  The only remaining risk is that, when the computed area
		//	happens to have integer dimensions, the inherent error
		//	in the floating-point value of contentScaleFactor might
		//	cause us to compute, say, a width of 636.999999999 instead
		//	of the required 637, which might force iOS to re-sample our image
		//	because 636.99999999 rounds down to 636, which is 1 less than
		//	the required 637.  To avoid that, we can ignore the floating-point
		//	value contentScaleFactor and instead take the content scale factor
		//	to be the ratio of two integers, namely the main screen's
		//
		//			width in native pixels
		//			----------------------
		//			   width in points
		//
		//	to ensure that integer results always come out exact.

		//	2020-12-11  My iPhone SE 2020 normally renders a 375×667 pt layout
		//	on its native 750×1334 px display, for an exact 2x scale factor.
		//	But it also offers a Display Zoom option, which treats
		//	that same native 750×1334 px display as 320×568 pt layout.
		//	The problem is that the scale factor is anisotropic:
		//
		//		the horizontal scale factor is  750/320 = 2.34375
		//		the  vertical  scale factor is 1334/568 = 2.3485915…
		//
		//	So some extra fussing around is required.

		//	The screen bounds (in points) are given
		//	relative to the current interface orientation.
		let theScreenSizeInPoints = UIScreen.main.bounds.size

		//	By contrast, the nativeBounds (in pixels) are always given
		//	in portrait orientation when running on an iOS device,
		//	but are given in the natural landscape orientation
		//	when running on a Mac.  So in effect at runtime
		//	we don't know whether the nativeBounds are using
		//	the same orientation as the regular "bounds",
		//	or the opposite.  So for simplicity let's just
		//	match long with long and short with short.
		let theScreenSizeInPixels: CGSize
		if ((UIScreen.main.nativeBounds.size.height >= UIScreen.main.nativeBounds.size.width)
		 == (theScreenSizeInPoints.height >= theScreenSizeInPoints.width)
		) {
			theScreenSizeInPixels = UIScreen.main.nativeBounds.size
		} else {
			//	Swap width and height.
			theScreenSizeInPixels = CGSize(
				width:  UIScreen.main.nativeBounds.size.height,
				height: UIScreen.main.nativeBounds.size.width)
		}
		
		//	By doing the multiplication before the division,
		//	we ensure that whenever the result should be an integer,
		//	it will get computed exactly.
		let theDrawableSize = CGSize(
			width:  (self.bounds.size.width  * theScreenSizeInPixels.width ) / theScreenSizeInPoints.width,
			height: (self.bounds.size.height * theScreenSizeInPixels.height) / theScreenSizeInPoints.height)

		if let theCAMetalLayer = self.layer as? CAMetalLayer {
			theCAMetalLayer.drawableSize = theDrawableSize
		}

		//	Request a redraw.
		itsDrawableSizeHasChanged = true
	}
	
	override func didMoveToWindow() {

		if self.window != nil {
		
			//	This GeometryGamesView has been added to a Window.
			
			//	Create itsAnimationTimer.
			//
			//		Note:  iOS will automatically pause the timer
			//		when our app is in the background.
			//
			let theAnimationTimer = CADisplayLink(
									target: self,
									selector: #selector(animationTimerFired))
			theAnimationTimer.add(to: RunLoop.current, forMode: RunLoop.Mode.default)
			itsAnimationTimer = theAnimationTimer

		} else {

			//	This GeometryGamesView has been removed from its Window.
			
			//	We must deallocate the CADisplayLink (itsAnimationTimer),
			//	because otherwise the run loop would keep a strong reference
			//	to the CADisplayLink, which keeps a strong reference
			//	to this GeometryGamesView, and thus this GeometryGamesView
			//	would never get deallocated.
			
			if let theAnimationTimer = itsAnimationTimer {	//	should never fail

				//	invalidate() clears the run loop's strong reference
				//	to the CADisplayLink, which deallocates the CADisplayLink
				//	and thus clears the CADisplayLink's strong reference
				//	to this GeometryGamesView.  This GeometryGamesView's
				//	instance variable itsAnimationTimer is a weak reference
				//	"upstream" to the CADisplayLink, so it doesn't prevent
				//	the CADisplayLink from getting deallocated.
				
				theAnimationTimer.invalidate()
			}
		}
	}


// MARK: -
// MARK: Animation

	@objc func animationTimerFired(displayLink: CADisplayLink) {

		//	Note:  iOS automatically stops the animation timer
		//	when the app is in the background, so we needn't worry
		//	about wasting time drawing a view that isn't visible.
		
		//	Update the model, noting the possibly new change count.
		let theModelChangeCount = itsModelData.updateModel()

		//	Has the model changed?
		let theModelHasChanged = (theModelChangeCount != itsPreviousModelChangeCount)
		itsPreviousModelChangeCount = theModelChangeCount
		
		//	Has the drawable size changed?
		let theDrawableSizeHasChanged = itsDrawableSizeHasChanged
		itsDrawableSizeHasChanged = false
		
		//	Redraw the view iff something has changed.
		if (theModelHasChanged || theDrawableSizeHasChanged) {
			redrawView()
		}

		//	Play any pending sound.
		playPendingSound()
	}

	func redrawView() {

		//	Wrap the render() call in an autorelease pool
		//	so that the drawable gets released as soon as
		//	the render() function returns.
		//
		autoreleasepool {

			if let theCAMetalLayer = self.layer as? CAMetalLayer {
			
				//	Ask the CAMetalLayer for the next available color buffer.
				//	nextDrawable() blocks until a drawable becomes available.
				//	It returns nil only if
				//		- theCAMetalLayer has an invalid combination of drawable properties,
				//		- all drawables are in use and the 1-second timeout has elapsed, or
				//		- the process is out of memory.
				//
				if let theDrawable = theCAMetalLayer.nextDrawable() {
					itsRenderer.render(
						modelData: itsModelData,
						drawable: theDrawable)
				}
			}
		}
	}
	

// MARK: -
// MARK: Gesture handling

	func gestureLocationAsDisplayPoint(_ cgPoint: CGPoint) -> DisplayPoint {

		//	Both the cgPoint and theViewBounds should be in points (not pixels),
		//	but either would be fine just so everything is consistent.

		let theViewBounds = self.bounds
		
		var theCGPoint = cgPoint

		//	theCGPoint in coordinates (0,0) to (width, height)
		theCGPoint.x -= theViewBounds.origin.x
		theCGPoint.y -= theViewBounds.origin.y

		//	Flip from iOS's top-down coordinates
		//	to our DisplayPoint's bottom-up coordinates.
		theCGPoint.y = theViewBounds.size.height - theCGPoint.y
		
		//	Package up the result.
		let theResult = DisplayPoint(
			x: theCGPoint.x,
			y: theCGPoint.y,
			viewWidth:  theViewBounds.size.width,
			viewHeight: theViewBounds.size.height)

		return theResult
	}

	func gestureTranslationAsDisplayPointMotion(_ translation: CGPoint) -> DisplayPointMotion {

		//	Both the translation and theViewBounds should be in points (not pixels),
		//	but either would be fine just so everything is consistent.

		let theViewBounds = self.bounds
		
		var theTranslation = translation

		//	Flip from iOS's top-down coordinates
		//	to our DisplayPointMotion's bottom-up coordinates.
		theTranslation.y = -theTranslation.y
		
		//	Package up the result.
		let theResult = DisplayPointMotion(
			deltaX: theTranslation.x,
			deltaY: theTranslation.y,
			viewWidth:  theViewBounds.size.width,
			viewHeight: theViewBounds.size.height)

		return theResult
	}
}
